from __future__ import annotations

from typing import Literal
from dataclasses import dataclass
from pathlib import Path

from elftools.elf.elffile import ELFFile

from .core import Core, RaspsimException
from .elf import ELF

class Raspsim(Core):
    def __init__(self):
        super().__init__()
        self._current_elf: ELF | None = None

    def load_elf(self, elf: ELF, abi: Literal["sysv"] = "sysv") -> None:
        """
        Populate the simulator with the ELF binary data.

        Args:
            sim: Raspsim instance.
            elf: ELF binary data.
            abi: ABI to use. Currently only supports sysv.
        """
        del abi  # unused

        self.registers.rip = elf.entry

        if elf.stack is None:
            raise RaspsimException("stack must be set")
        self.registers.rsp = elf.stack

        for segment in elf.segments:
            self.memmap(
                segment.vaddr, segment.prot, length=segment.size, data=segment.data or b""
            )
    
    def run(self) -> None:
        """
        Run the simulator.
        """
        try:
            super().run()
        except RaspsimException as e:
            path: str | None = None
            lineno: int | None = None
            line: str | None = None
            if self._current_elf is not None:
                path, lineno = self._current_elf.loc(self.registers.rip)
                try:
                    lines = Path(path).read_text().splitlines()
                    line = lines[lineno - 1]
                except Exception as ex:
                    line = "Internal error: " + str(ex)
            
            setattr(e, "file", path)
            setattr(e, "lineno", lineno)
            setattr(e, "line", line)
            raise e


"""
This module provides utilities for compiling code to ELF binaries, handling ELF files, and populating a simulator with ELF binary data.

Classes:
    ELF: Represents ELF binary data.
    Segment: Represents a memory segment.
    InvalidELFException: Exception raised for invalid ELF files.

Functions:
    rscompile: Context manager to compile code to an ELF binary.
    elf_add_trampoline: Adds a trampoline to the ELF binary.
    asm_preamble: Generates an assembly preamble.
    asm_stop_sim: Generates an assembly instruction to stop the simulator.
    populate_from_elf: Populates the simulator with ELF binary data.

Most of the code is taken from work done by Alexis Engelke for the rasptest project.
"""

import subprocess
import tempfile
from typing import Literal, IO, Generator, cast
from dataclasses import dataclass, field
from contextlib import contextmanager
import logging
from pathlib import Path
from inspect import cleandoc


from .core import RaspsimException, Prot, getProtFromELFSegment


@contextmanager
def rscompile(
    *,
    code: str | None = None,
    file: str | Path | None = None,
    entry_label: str = "_start",
    lang: Literal["c", "assembler"] = "assembler",
    compiler: str = "cc",
    std: str = "c17",
    text_segment_addr: int = 0x4000,
    additional_flags: list[str] | None = None,
    debug_symbols: bool = False,
) -> Generator[IO[bytes], None, None]:
    """
    Compile Code to an ELF binary.

    Args:
        code: Code to compile.
        file: File containing the code to compile.
        entry_label: Entry point label.
        lang: Language of the code.
        compiler: Compiler to use.
        std: Standard to use.
        text_segment_addr: Address where the text segment should be placed.
        additional_flags: Additional flags to pass to the compiler.
        debug_symbols: Whether to include debug symbols in the binary.

    Yields:
        A file-like object containing the compiled ELF binary.
    """
    if code is None and file is None:
        raise ValueError("Either code or file must be provided")

    if code is not None and file is not None:
        raise ValueError("Only one of code or file must be provided")

    args = [
        compiler,
        f"-x{lang}",
        "-pipe",
        f"-std={std}",
        "-static",
        "-march=k8",
        "-mno-mmx",
        "-mno-80387",
        "-mtune=generic",  # prevent excessive loop unrolling
        "-ffreestanding",
        "-nostdlib",
        "-D_MM_MALLOC_H_INCLUDED",
        "-fcf-protection=none",
        # Assembler arguments
        "-Wa,-march=k8+cmov+nommx",
        "-Wa,-mx86-used-note=no",
        "-Wa,--fatal-warnings",
        # Linker arguments
        "-Wl,-n",
        "-Wl,--fatal-warnings",  # fail if entry point doesn't exist
        "-Wl,--no-dynamic-linker",
        "-Wl,--build-id=none",
        "-Wl,-z,defs",
        "-Wl,-z,noexecstack",
        "-Wl,-z,norelro",
        "-Wl,-z,noseparate-code",
        f"-Wl,-Ttext={text_segment_addr:#x}",
        f"-Wl,-e,{entry_label}",
    ]
    if debug_symbols:
        args += ["-g", "-gdwarf-4"]
    args += additional_flags or []

    with tempfile.NamedTemporaryFile("w+b") as f:
        args += ["-o", f.name]

        if code is not None:
            subprocess.run(args + ["-"], input=code.encode(), check=True)
        else:
            if not Path(cast(str | Path, file)).exists():
                raise FileNotFoundError(f"File not found: {file}")
            subprocess.run(args + [str(file)], check=True)
        f.seek(0)
        yield f


def asm_preamble(label: str = "_start") -> str:
    return cleandoc(
        f"""
                    .intel_syntax noprefix
                    .global {label}
                    .section .text
                    {label}:
                    """
    )


def asm_stop_sim() -> str:
    return "int 0x80\n"





import sys
import argparse

from .utils import rscompile
from .raspsim import Raspsim
from .core import RaspsimException
from .elf import ELF


def main() -> None:
    parser = argparse.ArgumentParser(description="Raspsim simulator")
    parser.add_argument("file", type=argparse.FileType("r"), help="Source file", nargs="?")
    parser.add_argument("--debug", action="store_true", help="Enable debug mode")
    parser.add_argument(
        "-x",
        "--lang",
        choices=["c", "assembler"],
        default="assembler",
        help="Language of the source file",
    )
    parser.add_argument(
        "--entry-label",
        default="_start",
        help="Entry point label",
    )
    parser.add_argument(
        "--compiler",
        default="cc",
        help="Compiler to use",
    )
    parser.add_argument(
        "--std",
        default="c17",
        help="Standard to use",
    )

    args = parser.parse_args()

    if args.file:
        program = args.file.read()
    else:
        program = "\n".join(sys.stdin.readlines())

    with rscompile(code=program, entry_label=args.entry_label, lang=args.lang, compiler=args.compiler, std=args.std, debug_symbols=True) as f:
        elf = ELF.from_file(f)

    # Create a new Raspsim instance
    sim = Raspsim()
    sim.load_elf(elf)

    # Run the simulation
    try:
        sim.run()

        # Print the simulation state
        print(sim)
    except RaspsimException as e:
        if hasattr(e, "file") and hasattr(e, "lineno") and hasattr(e, "line"):
            print(f"Error in {e.file}:{e.lineno}: {e}", file=sys.stderr)
            raise e



if __name__ == "__main__":
    main()



from __future__ import annotations
import logging
from dataclasses import dataclass, field
from typing import IO

from elftools.elf.elffile import ELFFile

from .core import Prot, getProtFromELFSegment, RaspsimException

@dataclass
class Segment:
    """
    A class representing a memory segment.
    Attributes:
        vaddr (int): The virtual address of the segment.
        prot (Prot): The protection attributes of the segment.
        size (int): The size of the segment.
        data (bytes | None): Optional data contained in the segment.
    Methods:
        __repr__(): Returns a string representation of the Segment instance.
    Note:
        Length of `data` shall not exceed the size of the segment.
        If size is greater than the length of the data, the rest of the segment denotes zeroed memory.
    """

    vaddr: int
    prot: Prot
    size: int
    data: bytes | None = None
    stack: bool = False

    def __repr__(self):
        data_repr = f" data={self.data!r}" if self.data else ""
        return f"Segment(vaddr={self.vaddr:#x}, prot={self.prot}, size={self.size:#x}{data_repr})"


class InvalidELFException(Exception):
    """
    Exception raised when an invalid ELF file is encountered.
    """


@dataclass
class ELF:
    """
    ELF binary data.

    Attributes:
        entry (int): The entry point of the ELF binary.
        stack (int | None): The stack pointer, if available.
        segments (list[Segment]): A list of segments loaded into memory.
    Methods:
        __repr__(): Returns a string representation of the ELF object.
    Note:
        Only contains the segments that are loaded into memory. Requires x86_64 executable ELF.
    """

    entry: int
    stack: int | None = None
    segments: list[Segment] = field(default_factory=list)
    line_map: list[tuple[int, str, int]] = field(default_factory=list)

    def loc(self, addr: int) -> tuple[str, int]:
        """
        Get the filename and line number of the address.

        Args:
            addr (int): The address to look up.

        Returns:
            tuple[str, int]: A tuple containing the filename and line number.
        """
        for a, filename, line in self.line_map:
            if a < addr:
                return filename, line
        
        return "<unknown>", 0

    def __repr__(self):
        line_map_repr = ""
        if self.line_map:
            line_map_repr = (
                ", line_map=["
                + ", ".join(
                    f"({hex(addr)}, {filename}:{line})"
                    for addr, filename, line in self.line_map
                )
                + "]"
            )

        return f"ELF(entry={self.entry:#x}, stack={self.stack:#x}, segments={self.segments}{line_map_repr})"

    @staticmethod
    def _extract_linemap(elffile):
        if not elffile.has_dwarf_info():
            return []

        linemap: list[tuple[int, str, int]] = []
        dwarfinfo = elffile.get_dwarf_info()
        for cu in dwarfinfo.iter_CUs():
            line_program = dwarfinfo.line_program_for_CU(cu)
            if not line_program:
                continue
            file_entries = line_program.header["file_entry"]
            for lpe in line_program.get_entries():
                if not lpe.state:
                    continue
                if lpe.state.file:
                    filename = file_entries[lpe.state.file - 1].name.decode()
                else:
                    filename = "<unknown>"
                if linemap and linemap[-1][0] == lpe.state.address:
                    continue
                linemap.append((lpe.state.address, filename, lpe.state.line))
        
        linemap = sorted(linemap, key=lambda x: x[0])
        linemap.reverse()
        return linemap

    @classmethod
    def from_file(
        cls,
        f: IO[bytes],
        gnu_stack_size: int = 0x1000,
        gnu_stack_addr: int = 0x7FE000000000,
    ) -> ELF:
        """
        Load an ELF file and parse its segments.

        Args:
            elf (IO[bytes]): A byte stream representing the ELF file.
            gnu_stack_size (int, optional): The size of the stack when a GNU Stack Segment is found. Defaults to 0x1000.
            gnu_stack_addr (int, optional): The address of the segment when a GNU Stack Segment is found. Defaults to 0x7FE000000000.

        Returns:
            ELF: An ELF object containing the parsed segments and entry point.

        Raises:
            InvalidELFException: If the ELF file is not of type ET_EXEC or machine type EM_X86_64.
            InvalidELFException: If a segment's file size is greater than its memory size.
            ValueError: If the segment data length does not match the segment's file size.
            InvalidELFException: If an unhandled segment type is encountered.
        """
        # new stream from bytes
        elf = ELFFile(f)

        if elf["e_type"] != "ET_EXEC":
            raise InvalidELFException(f"Invalid ELF type {elf['e_type']}")

        if elf["e_machine"] != "EM_X86_64":
            raise InvalidELFException(f"Invalid ELF machine {elf['e_machine']}")

        loaded_elf = cls(entry=elf["e_entry"])

        for segment in elf.iter_segments():
            logging.debug(
                "[.] ELF segment: %s V=%#x FSZ=%#x MSZ=%#x flags=%d",
                segment["p_type"],
                segment["p_vaddr"],
                segment["p_filesz"],
                segment["p_memsz"],
                segment["p_flags"],
            )
            if segment["p_type"] == "PT_LOAD":
                if segment["p_filesz"] > segment["p_memsz"]:
                    raise InvalidELFException(
                        "Invalid ELF: corrupt binary. p_filesz > p_memsz"
                    )

                # Try to read segment data
                segment_data = segment.data()
                if len(segment_data) != segment["p_filesz"]:
                    raise ValueError("Invalid ELF: corrupt binary")

                logging.debug("[.] Read %d bytes of code/data", len(segment_data))
                loaded_elf.segments.append(
                    Segment(
                        vaddr=segment["p_vaddr"],
                        size=segment["p_memsz"],
                        data=segment_data,
                        prot=getProtFromELFSegment(segment["p_flags"]),
                    )
                )
            elif segment["p_type"] == "PT_GNU_STACK":
                loaded_elf.stack = gnu_stack_addr
                loaded_elf.segments.append(
                    Segment(
                        vaddr=gnu_stack_addr - gnu_stack_size,
                        size=gnu_stack_size,
                        prot=Prot.RW,
                        stack=True,
                    )
                )
            elif segment["p_type"] == "PT_NULL":
                pass
            else:
                raise InvalidELFException(
                    f"Invalid ELF: unhandled segment {segment['p_type']}"
                )

        loaded_elf.line_map = ELF._extract_linemap(elf)

        return loaded_elf

    

    def add_trampoline(self):
        """
        Adds a trampoline to the ELF binary.

        The trampoline is a call instruction to the entry point followed by an terminate instruction.
        If reaching the terminator indicates that the program has gracefully returned from the call.

        Parameters:
            elf (ELF): The ELF object to which the trampoline will be added.
        Raises:
            RaspsimException: If the entry point is not page-aligned or if the entry
                              point is less than 0x3000.

        Modifies:
            - Adjusts the entry point of the ELF object.
            - Appends a new segment containing the trampoline code to the ELF object's segments.
        """
        if self.entry & 0xFFF:
            raise RaspsimException("code must be page-aligned")

        if self.entry < 0x3000:
            raise RaspsimException("vaddr must be >= 0x3000 to use trampoline")

        trampoline_bias = 0x2000
        addr_start = max(0x1000, self.entry - trampoline_bias)
        reloff = self.entry - addr_start - 5
        trampoline = b"\xe8" + reloff.to_bytes(4, "little") + b"\xcd\x80"
        self.segments.append(
            Segment(vaddr=addr_start, prot=Prot.RX, size=len(trampoline), data=trampoline)
        )
        self.entry = addr_start


from __future__ import annotations

class iNMeta(type):
    """
    A metaclass for creating N-bit integer classes.
    """
    def __new__(mcs, name, bases, dct, n):
        dct['n'] = n
        return super().__new__(mcs, name, bases, dct)

    def __init__(cls, name, bases, dct, n):
        super().__init__(name, bases, dct)


class iN(metaclass=iNMeta, n=0):
    """
    A class representing an N-bit integer.
    """
    n: int

    def __init__(self, value: int | None = None):
        if value is None:
            value = 0

        self.value = value % (2**self.n)

    def __repr__(self):
        negrep = ""
        if self.value >> (self.n - 1):
            negrep = ", -" + str(2**self.n - self.value)

        return f"i{self.n}({self.value}, {self.value:#x}{negrep})"

    def __add__(self, other: iN | int):
        if isinstance(other, int):
            return self.__class__(self.value + other)
        return self.__class__(self.value + other.value)

    def __sub__(self, other: iN | int):
        if isinstance(other, int):
            return self.__class__(self.value - other)
        return self.__class__(self.value - other.value)

    def __mul__(self, other: iN):
        return self.__class__(self.value * other.value)

    def __truediv__(self, other: iN):
        return self.__class__(self.value // other.value)

    def __floordiv__(self, other: iN):
        return self.__class__(self.value // other.value)

    def __mod__(self, other: iN):
        return self.__class__(self.value % other.value)

    def __pow__(self, other: iN):
        return self.__class__(self.value**other.value)

    def __lshift__(self, other: iN):
        return self.__class__(self.value << other.value)

    def __rshift__(self, other: iN):
        return self.__class__(self.value >> other.value)

    def __and__(self, other: iN):
        return self.__class__(self.value & other.value)

    def __or__(self, other: iN):
        return self.__class__(self.value | other.value)

    def __xor__(self, other: iN):
        return self.__class__(self.value ^ other.value)

    def __neg__(self):
        return self.__class__(-self.value)

    def __pos__(self):
        return self.__class__(+self.value)

    def __invert__(self):
        return self.__class__(~self.value)

    def __eq__(self, other: object):
        if isinstance(other, int):
            return self.value == other
        elif isinstance(other, iN):
            return self.value == other.value
        raise TypeError(f"Cannot compare i{self.n} with {type(other)}")

    def __ne__(self, other: object):
        if isinstance(other, int):
            return self.value != other
        elif isinstance(other, iN):
            return self.value != other.value
        raise TypeError(f"Cannot compare i{self.n} with {type(other)}")

    def __lt__(self, other: object):
        if isinstance(other, int):
            return self.value < other
        elif isinstance(other, iN):
            return self.value < other.value
        raise TypeError(f"Cannot compare i{self.n} with {type(other)}")

    def __le__(self, other: object):
        if isinstance(other, int):
            return self.value <= other
        elif isinstance(other, iN):
            return self.value <= other.value
        raise TypeError(f"Cannot compare i{self.n} with {type(other)}")

    def __gt__(self, other: object):
        if isinstance(other, int):
            return self.value > other
        elif isinstance(other, iN):
            return self.value > other.value
        raise TypeError(f"Cannot compare i{self.n} with {type(other)}")

    def __ge__(self, other: object):
        if isinstance(other, int):
            return self.value >= other
        elif isinstance(other, iN):
            return self.value >= other.value
        raise TypeError(f"Cannot compare i{self.n} with {type(other)}")

    def __bool__(self):
        return bool(self.value)

    def __int__(self):
        return self.value


class i8(iN, metaclass=iNMeta, n=8):
    """
    A class representing an 8-bit integer.
    """


class i16(iN, metaclass=iNMeta, n=16):
    """
    A class representing a 16-bit integer.
    """


class i32(iN, metaclass=iNMeta, n=32):
    """
    A class representing a 32-bit integer.
    """


class i64(iN, metaclass=iNMeta, n=64):
    """
    A class representing a 64-bit integer.
    """